from __future__ import annotations

import io

from ormspace import model as md
from hx_markup import functions
from ormspace.keys import Key, TableKey
from starlette.requests import Request

from spacestar.component import init_element, Element


@md.modelmap
class SpaceModel(md.Model):
    
    @property
    def tablekey(self) -> str:
        return f'{self.table()}.{self.key}'
    
    @property
    def table_key(self) -> str:  #TODO: deletar
        return self.tablekey
    
    @classmethod
    def htmx(cls, **kwargs):
        return functions.join_htmx_attrs(**kwargs)
    
    @classmethod
    def field(cls, name: str):
        return cls.model_fields.get(name, None)
    
    async def display(self):
        with io.StringIO() as f:
            container: Element = init_element('div', id=self.table_key)
            container.children.append(init_element('h3', children=str(self)))
            container.children.append(init_element('ul', '.nav', children=[init_element('li','.nav-item', children=f'{k}: {v}') for k, v in dict(self).items()]))
            f.write(str(container))
            return f.getvalue()
        
    async def heading(self, tag: str, *args, **kwargs):
        with io.StringIO() as f:
            kwargs['children'] = str(self)
            f.write(str(init_element(tag, *args, **kwargs)))
            return f.getvalue()
        
    @classmethod
    def query_from_request(cls, request: Request):
        q, fields = {}, {**cls.model_fields}
        for k, v in request.query_params.items():
            q[f'{k}?contains'] = v.lower()
        return q
    
    @classmethod
    def query_from_dict(cls, data: dict):
        q, fields = {}, {**cls.model_fields}
        if data:
            for k, f in fields.items():
                if k in data.keys():
                    if f.annotation in [str, list[str]]:
                        q[f'{k}?contains'] = data[k]
        return q
        

        
    
    




from starlette.middleware import Middleware
from starlette.middleware.sessions import SessionMiddleware


from spacestar import spacestar_settings


session_middleware = Middleware(SessionMiddleware, secret_key=spacestar_settings.session_secret)


from __future__ import annotations

import io
import os
from abc import ABC, abstractmethod
from collections import defaultdict
from functools import wraps
from typing import Optional

from hx_markup import Element, functions
from lxml import etree
from lxml.builder import E
from ormspace.alias import QUERIES
from ormspace.model import getmodel
from starlette.datastructures import FormData
from starlette.requests import Request
from starlette.responses import HTMLResponse
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles

from spacestar.app import SpaceStar
from spacestar.model import SpaceModel


def add_staticfiles(_app: SpaceStar = None, *, path: str = '/static', directory: str = 'static'):
    def decorator(app):
        @wraps(app)
        def wrapper():
            app.routes.append(
                    Mount(path, app=StaticFiles(directory=os.path.join(os.getcwd(), directory)), name='static'))
            return app
        
        return wrapper()
    
    if _app is None:
        return decorator
    else:
        return decorator(_app)


class ResponseEngine(ABC):
    def __init__(self, request: Request, *args, **kwargs):
        self.request = request
        self.app: SpaceStar = self.request.app
        self.template_path = kwargs.pop('template_path', None)
        self.source = kwargs.pop('source', None)
        self.args = args
        self.kwargs = kwargs
        
    @property
    def engine(self):
        if self.template_path and self.app.templates_directory:
            return self.app.templates.get_template(self.template_path)
        elif self.source:
            return self.app.from_string(self.source)
        return self.app.index
    
    @abstractmethod
    async def data(self) -> dict:
        raise NotImplementedError
    
    async def run(self):
        return self.engine.render(self.request, **await self.data())
    
    # @abstractmethod
    # async def element(self):...
    #
    # async def html(self) -> str:
    #     return etree.tounicode(await self.element(), pretty_print=True)
    #
        
class ModelResponse(ResponseEngine):
    def __init__(self, request: Request, *args, **kwargs):
        super().__init__(request, *args, **kwargs)
        self.model = getmodel(self.request.path_params.get('item_name'))

    async def update_dependencies(self, lazy=False, queries: QUERIES | None = None):
        await self.model.update_dependencies_context(queries=queries, lazy=lazy)
    
    @property
    def fields(self):
        return self.model.model_fields.values()
    
    @property
    def query(self):
        result = self.model.query_from_request(request=self.request)
        return result
    
    @property
    def path(self):
        return self.request.url.path
    
    @property
    def field_names(self):
        return self.model.model_fields.keys()
    
    async def instances(self, lazy=False):
        return await self.model.sorted_instances_list(query=self.query, lazy=lazy)

    async def run(self):
        if self.template_path:
            return HTMLResponse(self.app.render(self.request, template=self.template_path, **await self.data()))
        elif self.source:
            return HTMLResponse(self.app.templates.from_string(self.source).render(request=self.request, **await self.data()))
        return self.app.response(self.request, **await self.data())
    
    @classmethod
    async def form_data_dict(cls, request):
        form_data = await request.form()
        base = defaultdict(list)
        for key, value in form_data.items():
            base[key].append(value)
        result = {}
        for key, value in base.items():
            if len(value) == 1:
                result[key] = value[0]
            elif len(value) > 1:
                result[key] = value
        return result
    
def wrap_header(*args, **kwargs):
    return str(Element('div', '#header', children=Element(*args, **kwargs)))
    
class ListResponse(ModelResponse):
    
    async def data(self) -> dict:
        return {
                'header': wrap_header('div', '.container-fluid', children=[Element('h1', children=self.app.title)]),
                'main': etree.tounicode(
                        E.div(
                                E.h2(f'lista de {self.model.plural()}'),
                                E.ul(*[E.li(str(i), {'class': 'list-group-item text-white'}) for i in
                                       await self.instances()],
                                     {'class': 'list-group', 'style': 'overflow-y: auto; max-height: 80%;'})
                        
                        )
                        
                ),
                'footer': etree.tounicode(E.footer(f'resultados para {functions.write_args(self.query.values())}', id='footer'))
        }
    
    # async def element(self) -> etree.Element:
    #     page = E.div(
    #                     E.h4('lista de {}'.format(self.model.plural()).title()),
    #                     E.ul(
    #                             *[E.li(str(i), {'class': 'list-group-item',
    #                                             'style': 'background-color: transparent; color: white'}) for i in
    #                               await self.model.sorted_instances_list(lazy=False, query=self.model.query_from_request(self.request))],
    #                             id='{}-list'.format(self.model.item_name()),
    #                             **{'class': 'list-group'},
    #                             style='overflow-y: auto; max-height: 75vh;'
    #                     ),
    #                     style='display: grid; justify-content: center;'
    #             )
    #     return page
        

class SearchResponse(ModelResponse):
    async def element(self):
        page = E.div(
                E.h4('resultado de pesquisa de {}'.format(self.model).title()),
        )
        return page
    

def list_component_route(_model: type[SpaceModel] = None):

    def wrapper(model: type[SpaceModel]):
        @wraps(model)
        def wrapped():
            async def list_component_endpoint(request: Request):
                return request.app.response(request,
                                            model=model,
                                            instances=model.sorted_instances_list(query={**request.query_params}),
                                            template=f'component/list.html')
            return list_component_endpoint
        model._list_component_route = wrapped()
        return model
    if _model:
        return wrapper(_model)
    return wrapper
                # with io.StringIO() as f:
                #     container = init_element('div', f'#{model.item_name()}__list__container .card-box')
                #     container.children.append(init_element('h3', children=f'Lista de {model.plural()}'))
                #     group = init_element('ul', 'nav')
                #     for item in items:
                #         group.children.append(init_element('li', f'#{model.item_name()}__{item.key} .nav-item'))
                #     container.children.append(group)
                #     f.write(str(container))
                #     text = f.getvalue()


import os
from secrets import token_hex

from hx_markup.element import NodeText, Render
from lxml.builder import E

from spacestar.settings import SpaceStarSettings


BOOTSTRAP_LINK: str = ('<link href="https://cdn.jsdelivr.net/npm/bootstrap@5.3.2/dist/css/bootstrap.min.css" '
                       'rel="stylesheet" integrity="sha384-T3c6CoIi6uLrA9TneNEoa7RxnatzjcDSCmG1MXxSR1GAsXEV/Dwwykc2MPK'
                       '8M2HN" crossorigin="anonymous">')
BOOTSTRAP_SCRIPT: str = ('<script src="https://cdn.jsdelivr.net/npm/bootstrap@5.3.2/dist/js/bootstrap.bundle.min.js" int'
                         'egrity="sha384-C6RzsynM9kWDrMNeT87bh95OGNyZPhcTNXj1NW7RuBCsyN/o0jlpcV8Qyq46cDfL" crossorigin='
                         '"anonymous"></script>')
META_VIEWPORT: str = '<meta name="viewport" content="width=device-width, initial-scale=1">'
META_CHARSET: str = '<meta charset="utf-8">'
HEAD_STYLE_LINK: str = '<link rel="stylesheet" href="static/css/main.css">'
HEAD_SCRIPT: str = '<script src="static/js/main.js"></script>'

CREATE_SESSION_SECRET = True
CREATE_CSRF_SECRET = True


SCRIPT_BODY_BUBLES: str = (
        "<script type='text/javascript'>const body = document.querySelector('body');const circles = document.createEleme"
        "nt('div');circles.id = 'circles';for(let i = 34; i-=1; i>=0){const circle = document.createElement('div'); "
        "circle.className = 'circle'; circle.id = `circle-${i}`;circle.style.display = 'block';circle.style.position = "
        "'absolute';circle.style.footer = `50%`;circle.style.left = `50%`;circle.style.zIndex = `100`;circle.style."
        "transform = `scale(1)`;circle.style.opacity = `1`;circles.append(circle)}; body.prepend(circles); body.onload "
        "= setupCircles; body.onresize = setupCircles; function getRandomInt(max) {return Math.floor(Math.random() * max"
        ");} function getRandomFloat(max) {return Math.random() * max;}; function changeCircle(item){item.style.footer "
        "= `${getRandomInt(50)}vh`;item.style.left = `${getRandomInt(100)}vw`;item.style.zIndex = `${-100 + getRandomInt"
        "(50)}`;item.style.transform = `scale(${getRandomInt(7)})`;item.style.opacity = `${getRandomFloat(.7)}`;}function"
        " setupCircles(){circles.childNodes.forEach(item=>changeCircle(item))}</script>"
)

CSS_HEAD_BUBLES = """<style>
@import url("https://cdn.jsdelivr.net/npm/bootstrap-icons@1.11.2/font/bootstrap-icons.min.css");
:root{
    --primary-color: #f77903;
    --secondary-color: #777;
    --dark-color: #333;
    --light-color: #ddd;
    --body-background: radial-gradient(
        circle at 100%,
        var(--dark-color),
        var(--dark-color) 50%,
        var(--light-color) 75%,
        var(--dark-color) 75%
        );
    }
html, body {padding: 0; margin: 0; box-sizing: border-box; min-height: 100vh; min-height: 100dvh;}
body {background: var(--body-background); background-repeat: no-repeat; > * {color: white;}}
legend, label, a {color: var(--primary-color);}
#main {background-color: rgba(0, 0, 0, .5); padding: clamp(1vw, 1rem, 5vw); max-height: 90dvh; overflow-y: auto; display: grid; justify-content: center; align-items: center; align-content: center;}
.darkorange {color: var(--primary-color);}
.darkorange:hover {color: white; text-shadow: var(--primary-color) .125rem .125rem 1rem; transition: all 300ms;}
.darkorange:focus{color: white; text-shadow: var(--primary-color) .125rem .125rem 1rem; border-block: var(--primary-color) solid .125rem; transition: all 300ms;}
.small-caps{font-variant: small-caps;}
#circles{position: absolute; z-index_from_string: -100; header: 0; left: 0; width: 100%; height: 100%}
.circle {width: 1vw; height: 1vw; footer: 0; left: 50%; border-radius: 50%; background-image: radial-gradient(circle at 50%, var(--light-color), var(--primary-color)); transition: all 1s ease-in-out; display: none;}
</style>
"""

if not os.getenv('SESSION_SECRET', None):
    with open(os.path.join(os.getcwd(), '.env'), 'r') as env_file:
        for line in env_file:
            if line.startswith('SESSION_SECRET'):
                CREATE_SESSION_SECRET = False
                break

if CREATE_SESSION_SECRET:
    with open(os.path.join(os.getcwd(), '.env'), 'a') as f:
        f.write(f'\nSESSION_SECRET={token_hex(32)}')
        
        
if not os.getenv('CSRF_SECRET', None):
    with open(os.path.join(os.getcwd(), '.env'), 'r') as env_file:
        for line in env_file:
            if line.startswith('CSRF_SECRET'):
                CREATE_CSRF_SECRET = False
                break

if CREATE_CSRF_SECRET:
    with open(os.path.join(os.getcwd(), '.env'), 'a') as f:
        f.write(f'\nCSRF_SECRET={token_hex(32)}')
        

spacestar_settings = SpaceStarSettings()


from __future__ import annotations

from hx_markup.element import Element
from spacestar import CSS_HEAD_BUBLES, SCRIPT_BODY_BUBLES


def init_element(tag, /, *args, **kwargs) -> Element:
    booleans = [*args]
    children = kwargs.pop('children', [])
    dataset = kwargs.pop('dataset', {})
    classlist = kwargs.pop('classlist', [])
    styles = kwargs.pop('styles', {})
    htmx = kwargs.pop('htmx', {})
    id = kwargs.pop('id', None)
    for item in booleans[:]:
        if item.startswith('#'):
            if not id:
                id = item[1:]
            booleans.remove(item)
        elif item.startswith('.'):
            classlist.append(item[1:])
            booleans.remove(item)
    return Element(tag, id, booleans, classlist, kwargs, htmx, dataset, styles, children=children)


bootstrap_sript = Element('script', src="https://cdn.jsdelivr.net/npm/bootstrap@5.3.2/dist/js/bootstrap.bundle.min.js",
        integrity="sha384-C6RzsynM9kWDrMNeT87bh95OGNyZPhcTNXj1NW7RuBCsyN/o0jlpcV8Qyq46cDfL",
        crossorigin="anonymous"
)

bootstrap_link = Element('link', href="https://cdn.jsdelivr.net/npm/bootstrap@5.3.2/dist/css/bootstrap.min.css",
        rel="stylesheet",
        integrity="sha384-T3c6CoIi6uLrA9TneNEoa7RxnatzjcDSCmG1MXxSR1GAsXEV/Dwwykc2MPK8M2HN",
        crossorigin="anonymous"
)


def page_head(title: str = None, children: list[Element | str] = None) -> Element:
    head = Element('head', children=[
            Element('meta', charset='utf-8'),
            Element('meta', content="width=device-width, initial-scale=1", name='viewport'),
            Element('title', children=title),
            bootstrap_link,
            CSS_HEAD_BUBLES
            # Element('style', children=[Render.selector_style('*', {'margin': '0', 'padding': '0'}), 'html {background-image: radial_gradient(x1:0, y1:0, x2:x=width, y2:y=height, padding: 0}']),
            # Element('script', src='/static/js/main.js'),
            # Element('link', rel='stylesheet', href='/static/css/main.css'),
    
    ])
    
    if isinstance(children, list):
        head.children.extend(children)
    elif isinstance(children, str):
        head.children.append(children)
    
    return head


def page_body(*args, **kwargs) -> Element:
    children = kwargs.pop('children', [])
    children.append(SCRIPT_BODY_BUBLES)
    return Element('body', *args, children=children, **kwargs)


def page_html(lang: str, title: str, children: list[Element | str] = None) -> Element:
    return Element('html', lang=lang, children=[
            page_head(title),
            page_body('#body', children=children or [Element('h1', children=title)])
    ], before='<!DOCTYPE html>')


def nav_list(children, /, *args, **kwargs) -> Element:
    return init_element('ul', '.navbar-nav', *args, children=children, **kwargs)
 
def nav_item(children: str | Element, *args, **kwargs) -> Element:
    return init_element('li', '.nav-item', *args, children=children, **kwargs)


def nav_link(children: str | Element, href: str, *args, **kwargs) -> Element:
    return init_element('a', '.nav-link', *args, href=href, children=children, **kwargs)


def full_nav_list(data: list[tuple[str, str]]) -> Element:
    links = [nav_link(*i) for i in data]
    return nav_list([nav_item(i) for i in links])


if __name__ == '__main__':
    print(page_html('en', 'spacestar').render())
    print(page_body('#body', children=[Element('h1', children='spacestar')]))


from __future__ import annotations

import os


def create_static_folder(path: str | os.PathLike = None) -> None:
    if not path:
        path = os.path.join(os.getcwd(), 'static')
    os.makedirs(path, exist_ok=True)
    os.makedirs(f'{path}/css', exist_ok=True)
    os.makedirs(f'{path}/js', exist_ok=True)

    with open(f'{path}/css/main.css', 'w') as f:
        f.write('* {margin: 0; padding: 0; box-sizing: border-box}\n')
    with open(f'{path}/js/main.js', 'w') as f:
        f.write('')


from typing import Optional

from ormspace.settings import SpaceSettings
from pydantic import ConfigDict


class SpaceStarSettings(SpaceSettings):
    session_secret: Optional[str] = None
    csrf_secret: Optional[str] = None


from __future__ import annotations

__ALL__ = ['BODY_SCRIPTS', 'HEAD_TEMPLATE', 'HEADER_TEMPLATE', 'FOOTER_TEMPLATE', 'INDEX_TEMPLATE', 'app_context']

import os
from typing import Any

from hx_markup import Element
from starlette.requests import Request

from spacestar import component as cp


def app_context(request: Request) -> dict[str, Any]:
    return {'app': request.app}

BODY_SCRIPTS = """
<script type="text/javascript">
function pixToFloat(string) {return Number.parseFloat(string.slice(0, -2))}
function pixToInt(string) {return Number.parseInt(string.slice(0, -2))}
function numberToPix(number) {return `${number}px`}
const getId = (id) => document.getElementById(id)

const body = getId('body');
const header = getId('header');
const main = getId('main');
const footer = getId('footer');

const circles = document.createElement('div');
circles.id = 'circles';
body.prepend(circles);
for(let i = 34; i-=1; i>=0){
    const circle = document.createElement('div');
    circle.className = 'circle';
    circle.id = `circle-${i}`;
    circle.style.display = 'block';
    circle.style.position = 'absolute';
    circle.style.bottom = `50%`;
    circle.style.left = `50%`;
    circle.style.zIndex = `100`;
    circle.style.transform = `scale(1)`;circle.style.opacity = `1`;
    circles.append(circle)
}
function getRandomInt(max) {return Math.floor(Math.random() * max);}
function getRandomFloat(max) {return Math.random() * max;}
function changeCircle(item){
    item.style.bottom = `${getRandomInt(50)}vh`;
    item.style.left = `${getRandomInt(100)}vw`;
    item.style.zIndex = `${-100 + getRandomInt(50)}`;
    item.style.transform = `scale(${getRandomInt(7)})`;
    item.style.opacity = `${getRandomFloat(.7)}`;
    }
function setupCircles(){circles.childNodes.forEach(item=>changeCircle(item))}
function setupPage(){
    const bodyHeight = getComputedStyle(body)['height'];
    const headerHeight = getComputedStyle(header)['height'];
    const footerHeight = getComputedStyle(footer)['height'];
    main.style.height = numberToPix(pixToFloat(bodyHeight) - pixToFloat(headerHeight) - pixToFloat(footerHeight));
        const wrapper = document.getElementById('wrapper');
    if(wrapper){
        wrapper.style.maxHeight = numberToPix(pixToFloat(main.style.height) - 30);
    }
    setupCircles();
}
body.onload = setupPage;
body.onresize = setupPage;
</script>"""

HEAD_TEMPLATE: str = """
<link crossorigin="anonymous" href="https://cdn.jsdelivr.net/npm/bootstrap@5.3.2/dist/css/bootstrap.min.css" integrity="sha384-T3c6CoIi6uLrA9TneNEoa7RxnatzjcDSCmG1MXxSR1GAsXEV/Dwwykc2MPK8M2HN" rel="stylesheet"/>
<style>
@import url("https://cdn.jsdelivr.net/npm/bootstrap-icons@1.11.2/font/bootstrap-icons.min.css");
:root{
    --primary-color: #f77903;
    --secondary-color: #777;
    --dark-color: #333;
    --light-color: #ddd;
    --body-background: radial-gradient(
        circle at 100%,
        var(--dark-color),
        var(--dark-color) 50%,
        var(--light-color) 75%,
        var(--dark-color) 75%
        );
    }
    * {padding: 0; margin: 0; box-sizing: border-box}
    html {overflow: hidden; height: 100vh; height: 100dvh}
    body {
        height: 100%;
        background-image: var(--body-background);
        background-repeat: no-repeat;
        background-position: center;
        background-size: cover;
        > * {color: white}
    }
    legend, label, a {color: var(--primary-color);}
    #header{background-color: rgba(0, 0, 0, .5);}
    #main {display: grid; justify-content: center; align-items: center; align-content: center;}
    #wrapper {overflow-y: auto;}
    #footer{background-color: rgba(0, 0, 0, .5);}
    #circles{position: absolute; z-index: -100; top: 0; left: 0; width: 100%; height: 100%}
    .list-group-item {background-color: transparent}
    .darkorange {color: var(--primary-color);}
    .darkorange:hover {color: white; text-shadow: var(--primary-color) .125rem .125rem 1rem; transition: all 300ms;}
    .darkorange:focus{color: white; text-shadow: var(--primary-color) .125rem .125rem 1rem; border-block: var(--primary-color) solid .125rem; transition: all 300ms;}
    .small-caps{font-variant: small-caps;}
    .circle {width: 1vw; height: 1vw; bottom: 0; left: 50%; border-radius: 50%; background-image: radial-gradient(circle at 50%, var(--light-color), var(--primary-color)); transition: all 500ms ease-in-out; display: none;}

</style>
<script src="https://unpkg.com/htmx.org@1.9.10" integrity="sha384-D1Kt99CQMDuVetoL1lrYwg5t+9QdHe7NLX/SoJYkXDFfX37iInKRy5xLSi8nO7UC" crossorigin="anonymous"></script>
"""

HEADER_TEMPLATE = """
<header id="header" class="px-2">
    <h1><a href="/" class="text-white" style="text-decoration: none">{}</a></h1>
</header>
"""

FOOTER_TEMPLATE = """
        <footer id="footer" class="fixed-bottom px-2">
	        <div class="container-fluid" style="display: grid; grid-template-columns: 1fr 3fr 1fr">
				<div class=""></div>
				<div class="small-caps text-center">Sistema de Prontuário Eletrônico onde os dados são apenas seus.</div>
				<div class=""></div>
			</div>
        </footer>
	"""

INDEX_TEMPLATE = """
<!DOCTYPE html>
<html lang="pt-BR">
    <head>
        <meta charset="utf-8"/>
        <meta content="width=device-width, initial-scale=1" name="viewport"/>
        <title>{{title or request.app.title}}</title>
        {{head or request.app.head}}
     </head>
     <body id="body">
        {{ header or request.app.header }}</header>
        <main id="main"><div id="wrapper" class="p-3 bg-dark bg-opacity-75">{{ main or request.app.main }}</div></main>
        {{ footer or request.app.footer }}
        {{ body_scripts or request.app.body_scripts }}
     </body>
</html>
"""


def create_templates_folder(path: str | os.PathLike = None, lang: str = None, title: str = None) -> None:
    templates_path = os.path.join(path or os.path.join(os.getcwd(), 'templates'))
    os.makedirs(templates_path, exist_ok=True)
    with open(f'{templates_path}/index.html', 'w') as f:
        f.write(str(cp.page_html(lang=lang or 'en', title=title or 'SpaceStar', children=[
                Element('h1', children=title or 'SpaceStar'),
                '{{content}}'
        ])))